// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_authentication_bloc/src/core/constants/form_field.dart';
import 'package:wyatt_authentication_bloc/src/domain/domain.dart';
import 'package:wyatt_authentication_bloc/src/features/sign_in/cubit/sign_in_cubit.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

mixin SignInWithEmailPassword<Data> on BaseSignInCubit<Data> {
  FutureOrResult<Data?> onSignInWithEmailAndPassword(
    Result<Account, AppException> result,
    WyattForm form,
  );

  void emailChanged(String value) {
    final emailValidatorType = formRepository
        .accessForm(formName)
        .validatorOf(AuthFormField.email)
        .runtimeType;
    assert(
      emailValidatorType == Email,
      'Use emailCustomChanged(...) with validator $emailValidatorType',
    );

    final Email email = Email.dirty(value);
    dataChanged(AuthFormField.email, email);
  }

  void passwordChanged(String value) {
    final passwordValidatorType = formRepository
        .accessForm(formName)
        .validatorOf(AuthFormField.password)
        .runtimeType;
    assert(
      passwordValidatorType == Password,
      'Use passwordCustomChanged(...) with validator $passwordValidatorType',
    );
    final Password password = Password.dirty(value);
    dataChanged(AuthFormField.password, password);
  }

  /// Same as [emailChanged] but with a custom [Validator].
  ///
  /// Sort of short hand for [dataChanged].
  void emailCustomChanged<
      Validator extends FormInputValidator<String?, ValidationError>>(
    Validator validator,
  ) {
    dataChanged(AuthFormField.email, validator);
  }

  /// Same as [passwordChanged] but with a custom [Validator].
  ///
  /// Sort of short hand for [dataChanged].
  void passwordCustomChanged<
      Validator extends FormInputValidator<String?, ValidationError>>(
    Validator validator,
  ) {
    dataChanged(AuthFormField.password, validator);
  }

  FutureOr<void> signInWithEmailAndPassword() async {
    if (state.status.isSubmissionInProgress) {
      return;
    }

    if (!state.status.isValidated) {
      return;
    }

    final form = formRepository.accessForm(formName);
    emit(
      SignInState(
        form: form,
        status: FormStatus.submissionInProgress,
      ),
    );

    final email = form.valueOf<String?>(AuthFormField.email);
    final password = form.valueOf<String?>(AuthFormField.password);

    if (email.isNullOrEmpty || password.isNullOrEmpty) {
      emit(
        SignInState(
          form: form,
          errorMessage: 'An error occured while retrieving data from the form.',
          status: FormStatus.submissionFailure,
        ),
      );
    }

    final result = await authenticationRepository.signInWithEmailAndPassword(
      email: email!,
      password: password!,
    );

    // Custom routine
    final customRoutineResult = await onSignInWithEmailAndPassword(
      result,
      form,
    );
    if (customRoutineResult.isErr) {
      final error = customRoutineResult.err!;
      emit(
        SignInState(
          form: form,
          errorMessage: error.message,
          status: FormStatus.submissionFailure,
        ),
      );
      addError(error);
    }

    // Check result
    if (result.isErr) {
      final error = result.err!;
      emit(
        SignInState(
          form: form,
          errorMessage: error.message,
          status: FormStatus.submissionFailure,
        ),
      );
      addError(error);
    }

    final account = result.ok!;

    final signedInSession = SessionWrapper(
      event: SignedInEvent(account: account),
      session: Session<Data>(
        account: account,
        data: customRoutineResult.ok,
      ),
    );

    authenticationRepository.addSession(signedInSession);

    emit(
      result.fold(
        (value) =>
            SignInState(form: form, status: FormStatus.submissionSuccess),
        (error) => SignInState(
          form: form,
          errorMessage: error.message,
          status: FormStatus.submissionFailure,
        ),
      ),
    );
  }
}
